#  Copyright 2023 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.
from alamos import NOOP, Instrumentation
from freighter import Payload, UnaryClient

from synnax.exceptions import QueryError
from synnax.ranger.payload import RangeKey, RangePayload


class _SetActiveRequest(Payload):
    range: RangeKey


class _SetActiveResponse(Payload):
    pass


class _ClearActiveRequest(Payload):
    pass


class _RetrieveActiveRequest(Payload):
    pass


class _RetrieveActiveResponse(Payload):
    range: RangePayload | None = None


_SET_ENDPOINT = "/range/set-active"
_CLEAR_ENDPOINT = "/range/clear-active"
_RETRIEVE_ENDPOINT = "/range/retrieve-active"


class Active:
    def __init__(
        self,
        client: UnaryClient,
        instrumentation: Instrumentation = NOOP,
    ) -> None:
        self.__client = client
        self.instrumentation = instrumentation

    def set(self, key: RangeKey) -> None:
        self.__client.send(
            _SET_ENDPOINT, _SetActiveRequest(range=key), _SetActiveResponse
        )

    def clear(self) -> None:
        _, exc = self.__client.send(
            _CLEAR_ENDPOINT, _ClearActiveRequest(), _SetActiveResponse
        )
        if exc is not None:
            raise exc

    def retrieve(self) -> RangePayload | None:
        res, exc = self.__client.send(
            _RETRIEVE_ENDPOINT, _RetrieveActiveRequest(), _RetrieveActiveResponse
        )
        if isinstance(exc, QueryError):
            return None
        if exc is not None:
            raise exc
        return res.range



#  Copyright 2023 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

from pathlib import Path

from synnax.io.csv import CSVReader, CSVWriter
from synnax.io.protocol import FileWriter, RowFileReader

READERS: list[type[RowFileReader]] = [
    CSVReader,
]

WRITERS: list[type[FileWriter]] = [
    CSVWriter,
]


class IOFactory:
    """A registry for retrieving readers for different file types."""

    reader: list[type[RowFileReader]]
    writers: list[type[FileWriter]]

    def __init__(
        self,
        readers: list[type[RowFileReader]] = READERS,
        writers: list[type[FileWriter]] = WRITERS,
    ):
        self.reader = readers
        self.writers = writers

    def new_reader(self, path: Path | str) -> RowFileReader:
        path = path if isinstance(path, Path) else Path(path)
        if not path.exists():
            raise FileNotFoundError(f"File not found: {path}")

        if not path.is_file():
            raise IsADirectoryError(f"Path is a directory: {path}")

        for _Reader in self.reader:
            if _Reader.match(path):
                return _Reader(path)

        raise NotImplementedError(f"File type not supported: {path}")

    def open_writer(self, path: Path) -> FileWriter:
        if not path.parent.exists():
            raise FileNotFoundError(f"File not found: {path}")

        if not path.parent.is_dir():
            raise IsADirectoryError(f"Path is a directory: {path}")

        for Writer in self.writers:
            if Writer.match(path):
                return Writer(path)

        raise NotImplementedError(f"File type not supported: {path}")

    def extensions(self) -> list[str]:
        extensions = set()
        for reader in self.reader:
            extensions.update(reader.extensions())
        return list(extensions)


IO_FACTORY = IOFactory()



#  Copyright 2023 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

from alamos import Instrumentation, trace
from freighter import Payload, UnaryClient

from synnax.channel.payload import (
    ChannelKeys,
    ChannelNames,
    ChannelPayload,
    ChannelParams,
    normalize_channel_params
)


class _CreateRequest(Payload):
    channels: list[ChannelPayload]


_Response = _CreateRequest

_CHANNEL_CREATE_ENDPOINT = "/channel/create"
_CHANNEL_DELETE_ENDPOINT = "/channel/delete"


class _DeleteRequest(Payload):
    keys: ChannelKeys | None
    names: ChannelNames | None


class _DeleteResponse(Payload):
    ...


class ChannelWriter:
    __ENDPOINT = "/channel/create"
    __client: UnaryClient
    instrumentation: Instrumentation

    def __init__(
        self,
        client: UnaryClient,
        instrumentation: Instrumentation,
    ):
        self.__client = client
        self.instrumentation = instrumentation

    @trace("debug")
    def create(
        self,
        channels: list[ChannelPayload],
    ) -> list[ChannelPayload]:
        req = _CreateRequest(channels=channels)
        res, exc = self.__client.send(self.__ENDPOINT, req, _Response)
        if exc is not None:
            raise exc
        return res.channels

    @trace("debug")
    def delete(self, channels: ChannelParams) -> None:
        normal = normalize_channel_params(channels)
        req = _DeleteRequest(**{normal.variant: normal.params})
        res, exc = self.__client.send(_CHANNEL_DELETE_ENDPOINT, req, _DeleteResponse)
        if exc is not None:
            raise exc
        return res



#  Copyright 2023 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

from __future__ import annotations

from typing import Protocol, cast

from alamos import NOOP, Instrumentation, trace
from freighter import Payload, UnaryClient

from synnax.channel.payload import (
    ChannelKey,
    ChannelKeys,
    ChannelName,
    ChannelNames,
    ChannelParams,
    ChannelPayload,
    normalize_channel_params,
)
from synnax.exceptions import NotFoundError


class _Request(Payload):
    names: list[str] | None = None
    keys: list[int] | None = None
    leaseholder: int | None = None


class _Response(Payload):
    channels: list[ChannelPayload] = list()
    not_found: list[str] | None = list()


class ChannelRetriever(Protocol):
    """Protocol for retrieving channel payloads from the cluster."""

    def retrieve(self, params: ChannelParams) -> list[ChannelPayload]:
        ...


class ClusterChannelRetriever:
    __ENDPOINT = "/channel/retrieve"
    __client: UnaryClient
    instrumentation: Instrumentation

    def __init__(
        self,
        client: UnaryClient,
        instrumentation: Instrumentation = NOOP,
    ) -> None:
        self.__client = client
        self.instrumentation = instrumentation

    def _(self) -> ChannelRetriever:
        return self

    @trace("debug")
    def retrieve(self, params: ChannelParams) -> list[ChannelPayload]:
        normal = normalize_channel_params(params)
        if len(normal.params) == 0:
            return list()
        return self.__execute(
            _Request(**{normal.variant: normal.params})
        )  # type: ignore

    def __execute(
        self,
        req: _Request,
    ) -> list[ChannelPayload]:
        res, exc = self.__client.send(self.__ENDPOINT, req, _Response)
        if exc is not None:
            raise exc
        return res.channels


class CacheChannelRetriever:
    __retriever: ChannelRetriever
    __channels: dict[ChannelKey, ChannelPayload]
    __names_to_keys: dict[ChannelName, ChannelKey]
    instrumentation: Instrumentation

    def __init__(
        self,
        retriever: ChannelRetriever,
        instrumentation: Instrumentation,
    ) -> None:
        self.__channels = dict()
        self.__names_to_keys = dict()
        self.instrumentation = instrumentation
        self.__retriever = retriever

    def _(self) -> ChannelRetriever:
        return self

    def __get(self, param: ChannelKey | ChannelName) -> ChannelPayload | None:
        if isinstance(param, ChannelKey):
            return self.__channels.get(param)
        key = self.__names_to_keys.get(param)
        return None if key is None else self.__channels.get(key)

    def __set(self, channels: list[ChannelPayload]) -> None:
        for channel in channels:
            self.__channels[channel.key] = channel
            self.__names_to_keys[channel.name] = channel.key

    @trace("debug")
    def retrieve(self, params: ChannelParams) -> list[ChannelPayload]:
        normal = normalize_channel_params(params)
        results = list()
        to_retrieve: ChannelKeys | ChannelNames = list()  # type: ignore
        for p in normal.params:
            ch = self.__get(p)
            if ch is None:
                to_retrieve.append(p)  # type: ignore
            else:
                results.append(ch)

        if len(to_retrieve) == 0:
            return results

        retrieved = self.__retriever.retrieve(to_retrieve)
        self.__set(retrieved)
        results.extend(retrieved)
        return results


def retrieve_required(
    r: ChannelRetriever, params: ChannelParams
) -> list[ChannelPayload]:
    normal = normalize_channel_params(params)
    results = r.retrieve(params)
    not_found = list()
    for p in normal.params:
        ch = next((c for c in results if c.key == p or c.name == p), None)
        if ch is None:
            not_found.append(p)
    if len(not_found) > 0:
        raise NotFoundError(f"Could not find channels: {not_found}")
    return results



from threading import Thread
from asyncio import events, tasks


def _cancel_all_tasks(loop):
    to_cancel = tasks.all_tasks(loop)
    if not to_cancel:
        return

    for task in to_cancel:
        task.cancel()

    loop.run_until_complete(tasks.gather(*to_cancel, return_exceptions=True))

    for task in to_cancel:
        if task.cancelled():
            continue
        if task.exception() is not None:
            loop.call_exception_handler(
                {
                    "message": "unhandled exception during asyncio.run() shutdown",
                    "exception": task.exception(),
                    "task": task,
                }
            )


class AsyncThread(Thread):
    def __init__(self):
        super().__init__()

    def run(self):
        loop = events.new_event_loop()
        try:
            events.set_event_loop(loop)
            loop.run_until_complete(self.run_async())
        finally:
            try:
                _cancel_all_tasks(loop)
                loop.run_until_complete(loop.shutdown_asyncgens())
                loop.run_until_complete(loop.shutdown_default_executor())
            finally:
                events.set_event_loop(None)
                loop.close()

    async def run_async(self):
        ...
